﻿#include "Display.h"

// Дисплейный буфер
volatile u8 Buffer[16];
// Раскладка сегментов
//
//    AAAA
//   F    B
//   F    B  DIV
//   F    B
//    GGGG
//   E    C  DIV
//   E    C
//   E    C
//    DDDD   DP
//
// Массив для перекодировки цифры/символа в набор сегментов
PROGMEM u8 Get_Segments[]=
{
	(1<<SEG_A)|(1<<SEG_B)|(1<<SEG_C)|(1<<SEG_D)|(1<<SEG_E)|(1<<SEG_F)|(0<<SEG_G),  // 0
	(1<<SEG_B)|(0<<SEG_B)|(1<<SEG_C)|(0<<SEG_D)|(0<<SEG_E)|(0<<SEG_F)|(0<<SEG_G),  // 1
	(1<<SEG_A)|(1<<SEG_B)|(0<<SEG_C)|(1<<SEG_D)|(1<<SEG_E)|(0<<SEG_F)|(1<<SEG_G),  // 2
	(1<<SEG_A)|(1<<SEG_B)|(1<<SEG_C)|(1<<SEG_D)|(0<<SEG_E)|(0<<SEG_F)|(1<<SEG_G),  // 3
	(1<<SEG_B)|(0<<SEG_B)|(1<<SEG_C)|(0<<SEG_D)|(0<<SEG_E)|(1<<SEG_F)|(1<<SEG_G),  // 4
	(1<<SEG_A)|(0<<SEG_B)|(1<<SEG_C)|(1<<SEG_D)|(0<<SEG_E)|(1<<SEG_F)|(1<<SEG_G),  // 5
	(1<<SEG_A)|(0<<SEG_B)|(1<<SEG_C)|(1<<SEG_D)|(1<<SEG_E)|(1<<SEG_F)|(1<<SEG_G),  // 6
	(1<<SEG_A)|(1<<SEG_B)|(1<<SEG_C)|(0<<SEG_D)|(0<<SEG_E)|(1<<SEG_F)|(0<<SEG_G),  // 7
	(1<<SEG_A)|(1<<SEG_B)|(1<<SEG_C)|(1<<SEG_D)|(1<<SEG_E)|(1<<SEG_F)|(1<<SEG_G),  // 8
	(1<<SEG_A)|(1<<SEG_B)|(1<<SEG_C)|(1<<SEG_D)|(0<<SEG_E)|(1<<SEG_F)|(1<<SEG_G),  // 9
	(0<<SEG_A)|(0<<SEG_B)|(0<<SEG_C)|(0<<SEG_D)|(0<<SEG_E)|(0<<SEG_F)|(0<<SEG_G),  // пробел (код 10)
	(0<<SEG_A)|(0<<SEG_B)|(0<<SEG_C)|(0<<SEG_D)|(0<<SEG_E)|(0<<SEG_F)|(1<<SEG_G),  // - (код 11)
	(1<<SEG_A)|(1<<SEG_B)|(1<<SEG_F)|(0<<SEG_D)|(0<<SEG_E)|(0<<SEG_F)|(1<<SEG_G),  // градус (код 12)
	(1<<SEG_A)|(1<<SEG_B)|(1<<SEG_C)|(0<<SEG_D)|(1<<SEG_E)|(1<<SEG_F)|(1<<SEG_G),  // A (код 13)
	(0<<SEG_A)|(0<<SEG_B)|(0<<SEG_C)|(1<<SEG_D)|(1<<SEG_E)|(0<<SEG_F)|(1<<SEG_G),  // c (код 14)
	(0<<SEG_A)|(0<<SEG_B)|(0<<SEG_C)|(0<<SEG_D)|(1<<SEG_E)|(0<<SEG_F)|(1<<SEG_G)   // r (код 15)
};
// Массив, содержащий адреса дисплейного буфера для включения сегментов
// Данные по сегментам слева направо: A,B,C,D,E,F,G,DP,DIV
PROGMEM u8 Segment_Addr[]=
{
	0x0A, 0x00, 0x0A, 0x0B, 0x0B, 0x0B, 0x0A, 0x00, 0x00,   // Первое знакоместо
	0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 0x01, 0x01,
	0x0A, 0x0A, 0x0A, 0x02, 0x01, 0x02, 0x02, 0x09, 0x00,
	0x02, 0x09, 0x09, 0x02, 0x09, 0x09, 0x02, 0x08, 0x09,
	0x03, 0x08, 0x08, 0x03, 0x08, 0x08, 0x03, 0x07, 0x00,
	0x03, 0x07, 0x07, 0x03, 0x07, 0x07, 0x03, 0x06, 0x00,
	0x04, 0x06, 0x06, 0x04, 0x06, 0x06, 0x04, 0x05, 0x00,
	0x04, 0x05, 0x05, 0x04, 0x05, 0x05, 0x04, 0x00, 0x00    // Последнее знакоместо
};
// Массив, содержащий битовые маски дисплейного буфера для включения сегментов
// Данные по сегментам слева направо: A,B,C,D,E,F,G,DP,DIV
PROGMEM u8 Segment_Mask[]=
{
	0x10, 0x01, 0x40, 0x04, 0x02, 0x01, 0x20, 0x04, 0x00,   // Первое знакоместо
	0x01, 0x02, 0x04, 0x40, 0x02, 0x10, 0x20, 0x40, 0x10,
	0x01, 0x02, 0x04, 0x04, 0x20, 0x01, 0x02, 0x40, 0x00,
	0x10, 0x01, 0x02, 0x40, 0x20, 0x10, 0x20, 0x40, 0x04,
	0x01, 0x01, 0x02, 0x04, 0x20, 0x10, 0x02, 0x40, 0x00,
	0x10, 0x01, 0x02, 0x40, 0x20, 0x10, 0x20, 0x40, 0x04,
	0x01, 0x01, 0x02, 0x04, 0x20, 0x10, 0x02, 0x40, 0x00,
	0x10, 0x01, 0x02, 0x40, 0x20, 0x10, 0x20, 0x00, 0x00    // Последнее знакоместо
};


// Инициализирует дисплей
void Display_Init (void)
{
	// Источником тактирования дисплея выбран внутренний RC-генератор
	Display_Write_Cmd (CMD_SYS_CLK_RC_256K);
	// Выбор уровня напряжений и схемы коммутации сегментов
	Display_Write_Cmd (CMD_BIAS_1DIV3_COMMONS_3);
	// Включение тактирования дисплея
	Display_Write_Cmd (CMD_SYS_EN);
	// Выбор частоты пищалки 4 кГц
	Display_Write_Cmd (CMD_TONE_4K);
	// Очистка дисплейного буфера
	Display_Clear();
	// Отправка буфера на дисплей
	Display_Update();
	// Включение генератора напряжения дисплея
	Display_Write_Cmd (CMD_LCD_ON);
}


// Очистка дисплейного буфера
// После использования функции необходимо отправить буфер на дисплей
void Display_Clear (void)
{
	for (u8 i = 0; i < sizeof(Buffer); i++) Buffer[i] = 0;
}


// Тестовый режим
// Засвечивает все сегменты
// После использования функции необходимо отправить буфер на дисплей
void Display_Test (void)
{
	for (u8 i = 0; i < sizeof(Buffer); i++) Buffer[i] = 0xFF;
}


// Отправляет команду в дисплей
// Cmd - команда (9 бит)
void Display_Write_Cmd (u16 Cmd)
{
	// Выбор чипа, начало обмена данными
	Serial_Bus_Chip_Select();
	// Выбор режима 12 бит, старший бит первый
	Serial_Bus_Mode_Set (12, MSB_First);
	// Отправка префикса и команды
	Serial_Bus_Send_Data ((CMD_WRITE << 9) | Cmd);
	// Окончание обмена данными
	Serial_Bus_Chip_DeSelect();
}


// Отправляет 4 бита данных в дисплей
// Эквивалентно записи одного адреса
// Address - адрес от 0 до 31
// Data - данные (4 младших бита)
void Display_Write_Data_Single (u8 Address, u8 Data)
{
	// Выбор чипа, начало обмена данными
	Serial_Bus_Chip_Select();
	// Выбор режима 9 бит, старший бит первый
	Serial_Bus_Mode_Set (9, MSB_First);
	// Отправка префикса и адреса
	Serial_Bus_Send_Data ((DAT_WRITE << 6) | Address);
	// Выбор режима 4 бита, младший бит первый
	Serial_Bus_Mode_Set (4, LSB_First);
	// Отправка данных
	Serial_Bus_Send_Data (Data);
	// Окончание обмена данными
	Serial_Bus_Chip_DeSelect();
}


// Отправляет 8 бит данных в дисплей
// Эквивалентно записи двух адресов
// Address - адрес от 0 до 15
// Data - данные (8 бит)
void Display_Write_Data_Double (u8 Address, u8 Data)
{
	// Выбор чипа, начало обмена данными
	Serial_Bus_Chip_Select();
	// Выбор режима 9 бит, старший бит первый
	Serial_Bus_Mode_Set (9, MSB_First);
	// Отправка префикса и адреса
	Serial_Bus_Send_Data ((DAT_WRITE << 6) | (Address * 2));
	// Выбор режима 8 бит, младший бит первый
	Serial_Bus_Mode_Set (8, LSB_First);
	// Отправка данных
	Serial_Bus_Send_Data (Data);
	// Окончание обмена данными
	Serial_Bus_Chip_DeSelect();
}


// Отправляет 16 байт в ОЗУ дисплея
// Data - указатель на массив с данными (16 байт)
void Display_Write_Data_All (u8* Data)
{
	// Выбор чипа, начало обмена данными
	Serial_Bus_Chip_Select();
	// Выбор режима 9 бит, старший бит первый
	Serial_Bus_Mode_Set (9, MSB_First);
	// Отправка префикса и адреса
	Serial_Bus_Send_Data ((DAT_WRITE << 6) | 0);
	// Выбор режима 8 бит, младший бит первый
	Serial_Bus_Mode_Set (8, LSB_First);
	// Отправка данных из буфера на дисплей
	for (u8 i = 0; i < 16; i++)
	{
		Serial_Bus_Send_Data (*(Data+i));
	}
	// Окончание обмена данными
	Serial_Bus_Chip_DeSelect();
}


// Отправка буфера на дисплей
void Display_Update (void)
{
	Display_Write_Data_All (Buffer);
}


// Вывод символа в дисплейный буфер
// pos - позиция 0...7
// c - цифра (0...9) либо код символа (10 и более)
// После использования функции необходимо отправить буфер на дисплей
void Display_Sym (u8 pos, u8 c)
{
	// Защита от выхода за границы буфера
	if (pos >= POS_TOTAL) return;
	// Переменные для хранения адреса, маски, и номера сегмента
	u8 Adr;
	u8 Msk;
	u8 Seg_Offset;
	// Цикл прохода по всем сегментам
	for (Seg_Offset = SEG_A; Seg_Offset <= SEG_G; Seg_Offset++)
	{
		// Отображение цифры или символа
		if (pgm_read_byte(&Get_Segments[c]) & (1 << Seg_Offset))
		{
			// Вычисление адреса и маски
			Adr = pgm_read_byte(&Segment_Addr[(pos * SEG_TOTAL) + Seg_Offset]);
			Msk = pgm_read_byte(&Segment_Mask[(pos * SEG_TOTAL) + Seg_Offset]);
			// Запись в дисплейный буфер
			Buffer[Adr] |= Msk;
		}
	}
}


// Вывод точки/двоеточия в дисплейный буфер
// pos - позиция 0...7
// div - тип разделителя: точка (0) либо двоеточие (1)
// После использования функции необходимо отправить буфер на дисплей
void Display_Div (u8 pos, Display_Div_Type_t div)
{
	// Защита от выхода за границы буфера
	if (pos >= POS_TOTAL) return;
	// Переменные для хранения адреса, маски, и номера сегмента
	u8 Adr;
	u8 Msk;
	u8 Seg_Offset;
	// Выбор типа разделителя - точка или двоеточие
	if (div == Double) Seg_Offset = SEG_DIV; else Seg_Offset = SEG_DP;
	// Вычисление адреса и маски
	Adr = pgm_read_byte(&Segment_Addr[(pos * SEG_TOTAL) + Seg_Offset]);
	Msk = pgm_read_byte(&Segment_Mask[(pos * SEG_TOTAL) + Seg_Offset]);
	// Запись в дисплейный буфер
	Buffer[Adr] |= Msk;
}


// Вывод числа в дисплейный буфер
// Число беззнаковое 32-битное
// pos - позиция 0...7
// После использования функции необходимо отправить буфер на дисплей
void Display_Num (u8 pos, u32 n)
{
	// Указатель на строку
	char* Stroka;
	// Конвертируем число в строку, нацеливаем указатель на начало строки
	Stroka = utoa (n);
	// Выводим на дисплей
	while (*Stroka != 0)
	{
		// Конвертация и отображение числа
		Display_Sym (pos, (*Stroka) - '0');
		// Перемещаем указатель на следующий символ
		Stroka++;
		// Перемещаем указатель на следующее знакоместо дисплея
		pos++;
	}
}


// Вывод числа в дисплейный буфер
// Число знаковое 32-битное
// pos - позиция 0...7
// После использования функции необходимо отправить буфер на дисплей
void Display_SNum (u8 pos, s32 n)
{
	// Указатель на строку
	char* Stroka;
	// Конвертируем число в строку, нацеливаем указатель на начало строки
	Stroka = itoa (n);
	// Выводим на дисплей
	while (*Stroka != 0)
	{
		// Конвертация и отображение минуса
		if (*Stroka == '-') Display_Sym (pos, SYM_MINUS);
		// Конвертация и отображение числа
		else Display_Sym (pos, (*Stroka) - '0');
		// Перемещаем указатель на следующий символ
		Stroka++;
		// Перемещаем указатель на следующее знакоместо дисплея
		pos++;
	}
}


// Включить подсветку дисплея
void Display_Backlight_On (void)
{
	Display_Write_Cmd (CMD_IRQ_EN);
}


// Отключить подсветку дисплея
void Display_Backlight_Off (void)
{
	Display_Write_Cmd (CMD_IRQ_DIS);
}


// Включить пищалку
void Display_Buzzer_On (void)
{
	Display_Write_Cmd (CMD_TONE_ON);
}


// Отключить пищалку
void Display_Buzzer_Off (void)
{
	Display_Write_Cmd (CMD_TONE_OFF);
}

